#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "optimizer.h"

#include "codegen.h"
#include "tokenizer.h"

extern _codelist codelist;

extern _token tokens;

/*
long buildvarlist(long lLevel)
	{
	return 0;
	}

long copycode(_codelist *p, long lnItems, _duplicatecode *pDup)
	{
	if (!p)
		return 1;

	_codelist *pList = p;
	long i = 0;

	while (pList && i < lnItems)
		{
		pDup->lCodeArray[i++] = pList->lCode;
		pList = pList->pNext;
		}

	if (!pList && i < lnItems) //we can't get enough code
		return 2;

	return 0;
	}

long replacecode(_duplicatecode *pDup)
	{
	pDup->
	}

long optimizesize(long lLevel)
	{
	_codelist *p = &codelist;
	long i;

	_duplicatecode dup, dupcopy;

	dup.lnTimes = 0;
	dup.lSize = 0;
	dup.lYield = 0;
	bool bOK = true;

	//iterate through the code list and determine which de-inline gives us the best yield
	while (p)
		{
		for (i = 0; i < 0x100 && bOK; i++)
			{
			bOK = copycode(p, i) == 0;

			if (bOK)
				{
				dupcopy.lnTimes = counttimes(p, &dupcopy);
				dupcopy.lYield = (dupcopy.lnTimes * i) - i - 36; //subtract overhead to get the net yield
				dupcopy.lSize = i;
				dupcopy.pFirstOccurance = p;

				if (dupcopy.lYield > 0 && dupcopy.lYield > dup.lYield)
					{
					//if the new chunk has a yield greater than 0 and gives us a better yield than the old one, copy it into dup
					memcpy(dup.lCodeArray, dupcopy.lCodeArray, i);
					dup.lnTimes = dupcopy.lnTimes;
					dup.lSize = dupcopy.lSize;
					dup.lYield = dupcopy.lYield;
					dup.pFirstOccurance = dupcopy.pFirstOccurance;
					}
				}
			}

		p = p->pNext;
		}

	//we have the best code chunk now, so see if yield is greater than 0
	if (dup.lYield > 0)
		replacecode(&dup);

	return dup.lYield > 0;
	}*/

long optimizespeed_post(long lLevel)
	{
	long lShiftDivCount = 0;
	long lShiftMulCount = 0;

	_codelist *p  = &codelist;

	while (p)
		{
		if ((p->lCode == CODE_DIV) && (p->pPrev) && (p->pPrev->pPrev) && (p->pPrev->pPrev->lCode == CODE_PUSHCONSTANT))
			{
			//we found x / n, we can SAR
			if ((p->pPrev->lCode % 2) == 0)
				{
				long l = 0;
				while (pow(2, l) < p->pPrev->lCode)
					l++;

				p->pPrev->lCode = l;

				p->lCode = CODE_SAR;
				lShiftDivCount++;
				}
			}
		else if ((p->lCode == CODE_MUL) && (p->pPrev) && (p->pPrev->pPrev) && (p->pPrev->pPrev->lCode == CODE_PUSHCONSTANT))
			{
			//we found x * n, we can SHL

			if ((p->pPrev->lCode % 2) == 0)
				{
				long l = 0;
				while (pow(2, l) < p->pPrev->lCode)
					l++;

				p->lCode = CODE_SHL;
				lShiftMulCount++;
				}
			else if (p->pPrev->pPrev->pPrev && p->pPrev->pPrev->pPrev->pPrev && p->pPrev->pPrev->pPrev->pPrev->lCode == CODE_PUSHCONSTANT)
				{
				//we found n * x, swap n and x and change mul to shl. since both values are constants, we can just switch the values

				if ((p->pPrev->pPrev->pPrev->lCode % 2) == 0)
					{
					long l = 0;
					while (pow(2, l) < p->pPrev->pPrev->pPrev->lCode)
						l++;

					p->pPrev->pPrev->pPrev->lCode = p->pPrev->lCode;
					p->pPrev->lCode = l;
					p->lCode = CODE_SHL;
					lShiftMulCount++;
					}
				}
			}
		else if ((p->lCode == CODE_MUL) && (p->pPrev) && (p->pPrev->pPrev) && (p->pPrev->pPrev->lCode == CODE_PUSHVAR || p->pPrev->pPrev->lCode == CODE_PUSHSTATIC))
			{
			//for x * y, if y is a local or static var, check to see if x is a constant

			if (p->pPrev->pPrev->pPrev && p->pPrev->pPrev->pPrev->pPrev && p->pPrev->pPrev->pPrev->pPrev->lCode == CODE_PUSHCONSTANT)
				{
				//if so, see if we can switch the values

				if ((p->pPrev->pPrev->pPrev->lCode % 2) == 0)
					{
					long l = 0;

					while (pow(2, l) < p->pPrev->pPrev->pPrev->lCode)
						l++;

					p->pPrev->pPrev->pPrev->pPrev->lCode = p->pPrev->pPrev->lCode; //switch the push instructions
					p->pPrev->pPrev->lCode = CODE_PUSHCONSTANT;

					p->pPrev->pPrev->pPrev->lCode = p->pPrev->lCode; //switch the numbers
					p->pPrev->lCode = l;

					p->lCode = CODE_SHL; //change mul to shl
					}
				}
			}

		p = p->pNext;
		}

	printf("\tOptimization for speed results:\n");
	printf("\t\tChanged %d div instructions to sar\n\t\tChanged %d mul instructions to shl\n", lShiftDivCount, lShiftMulCount);
	return 0;
	}

long optimize_post(long lLevel, long lType)
	{
	//if (lType & 1)
	//	optimizesize(lLevel);

	if (lType & 1)
		{
		printf("\tOptimization for size results:\n");
		printf("\t\tNo size optimizations available\n");
		}

	if (lType & 2)
		optimizespeed_post(lLevel);

	return 0;
	}

long eval_const_expressions_pre(void)
	{
	_token *p = &tokens;

	bool bSwitchA = false, bSwitchB = false;
	double f1, f2;
	long lResult;

	long lBytes = 0;

	while (p && p->psz)
		{
		if ((p->pParent) && //';'
			(p->pParent->pParent) && //const
			(p->pParent->pParent->pParent) && (p->pParent->pParent->pParent->lType == PUSHCONSTANT) && //push const
			(p->pParent->pParent->pParent->pParent) && //';'
			(p->pParent->pParent->pParent->pParent->pParent) && //const
			(p->pParent->pParent->pParent->pParent->pParent->pParent) && (p->pParent->pParent->pParent->pParent->pParent->pParent->lType == PUSHCONSTANT))//push const
			{
			switch (p->lType)
				{
				case ADD:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = f1 + f2;
					bSwitchA = true;
					break;
				case SUB:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = f1 - f2;
					bSwitchA = true;
					break;
				case MUL:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = f1 * f2;
					bSwitchA = true;
					break;
				case DIV:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = f1 / f2;
					bSwitchA = true;
					break;
				case AND:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 & (long)f2;
					bSwitchA = true;
					break;
				case OR:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 | (long)f2;
					bSwitchA = true;
					break;
				case XOR:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 ^ (long)f2;
					bSwitchA = true;
					break;
				case SHL:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 << (long)f2;
					bSwitchA = true;
					break;
				case SAR:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 >> (long)f2;
					bSwitchA = true;
					break;
				case SETL:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 < (long)f2;
					bSwitchA = true;
					break;
				case SETLE:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 <= (long)f2;
					bSwitchA = true;
					break;
				case SETG:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 > (long)f2;
					bSwitchA = true;
					break;
				case SETGE:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 >= (long)f2;
					bSwitchA = true;
					break;
				case CMP:
				/*	f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 == (long)f2;
					break;*/
				case SETE:
					//fall through
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 == (long)f2;
					bSwitchA = true;
					break;
				case SETNE:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 != (long)f2;
					bSwitchA = true;
					break;
				case LAND:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 && (long)f2;
					bSwitchA = true;
					break;
				case LOR:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = (long)f1 || (long)f2;
					bSwitchA = true;
					break;
				case LXOR:
					f1 = atof(p->pParent->pParent->pParent->pParent->pParent->psz);
					f2 = atof(p->pParent->pParent->psz);
					lResult = ((long)f1 == 1) ^ ((long)f2 == 1);
					bSwitchA = true;
					break;
				default:
					break;
				}
			}

		else if ((p->pParent) && //';'
				(p->pParent->pParent) && //const
				(p->pParent->pParent->pParent) && (p->pParent->pParent->pParent->lType == PUSHCONSTANT)) //push const
			{
			switch (p->lType)
				{
				case NOT:
					f1 = atof(p->pParent->pParent->psz);
					lResult = ~((long)f1);
					bSwitchB = true;
					break;
				case NEG:
					f1 = atof(p->pParent->pParent->psz);
					lResult = !((long)f1);
					bSwitchB = true;
					break;
				default:
					break;
				}
			}

		if (bSwitchA)
			{
			lBytes += 12;

			bSwitchA = false;
			p->lType = INT;
			delete[] p->psz;
			char tempbuffer[0x20];
			strcpy(tempbuffer, ltoa(lResult, tempbuffer, 10));
			p->psz = new char [strlen(tempbuffer) + 1];
			strcpy(p->psz, tempbuffer);

			p->pParent->pParent->pParent->pParent->pParent->pParent->lType = NOCODE;
			p->pParent->pParent->pParent->pParent->pParent->lType = NOCODE;
			p->pParent->pParent->pParent->pParent->lType = NOCODE;
			p->pParent->pParent->pParent->lType = NOCODE;
			p->pParent->pParent->lType = NOCODE;
			p->pParent->lType = PUSHCONSTANT;

			p = &tokens;
			}
		else if (bSwitchB)
			{
			lBytes += 4;

			bSwitchB = false;
			p->lType = INT;
			delete[] p->psz;
			char tempbuffer[0x20];
			strcpy(tempbuffer, ltoa(lResult, tempbuffer, 10));
			p->psz = new char [strlen(tempbuffer) + 1];
			strcpy(p->psz, tempbuffer);

			p->pParent->lType = PUSHCONSTANT;
			p->pParent->pParent->lType = NOCODE;
			p->pParent->pParent->pParent->lType = NOCODE;

			p = &tokens;
			}
		else
			p = p->pNext;
		}

	return lBytes;
	}

long optimize_pre(long lLevel, long lType)
	{
	long lResult;
	if ((lType & 1) || (lType & 2))
		lResult = eval_const_expressions_pre();

	if (lType & 1) 
		printf("\tOptimiziation for size results:\n");
	if (lType & 2)
		printf("\tOptimization for speed results:\n");

	if ((lType & 1) || (lType & 2))
		printf("\t\tRemoved %d redundant bytes of code (speed and size)\n", lResult);

	return 0;
	}